"""
Utility functions for regular expressions. Builds on top of the standard library #re module.
"""

import re
import sys
import typing as t

if sys.version_info[:2] <= (3, 6):
    Pattern = t.Any
    Match = t.Any
else:
    Pattern = re.Pattern
    Match = re.Match


class MatchFullError(ValueError):
    """
    Raised when #match_full() cannot consume the full string.
    """

    def __init__(self, regex: "Pattern", string: str, endpos: int) -> None:
        self.regex = regex
        self.string = string
        self.endpos = endpos

    def __str__(self) -> str:
        return "could not consume whole string with regex {} (got until position {})".format(self.regex, self.endpos)


def matchiter(expr: t.Union[str, "Pattern"], string: str, flags: int = 0) -> t.Iterable["Match"]:
    """
    Like #re.finditer(), but uses #re.match() instead of #re.search().
    """

    if isinstance(expr, str):
        expr = re.compile(expr, flags)

    start = 0
    while True:
        match = expr.match(string, start)
        if not match:
            break
        start = match.end()
        yield match


def match_full(expr: t.Union[str, "Pattern"], string: str, flags: int = 0) -> t.Iterable["Match"]:
    """
    Like #matchiter(), but raises a #MatchAllError if the *expr* does not match any number of
    times over the entire string from start to finish.
    """

    if isinstance(expr, str):
        expr = re.compile(expr, flags)

    end = 0
    for match in matchiter(expr, string, flags):
        yield match
        end = match.end()

    if end != len(string):
        raise MatchFullError(expr, string, end)



import datetime
import typing as t
from dataclasses import dataclass, field

from .format import _datetime_format, date_format, datetime_format, time_format


def _formulate_parse_error(name: str, formats: t.Sequence[_datetime_format], s: str) -> ValueError:
    return ValueError(
        f'"{s}" does not match {name} date formats ({len(formats)}):'
        + "".join(f"\n  | {x.format_str}" for x in formats)
    )


@dataclass
class format_set:
    """
    Format sets represent a group of date, time and dateime formats. When formatting a value, it will
    use the first format defined in the group. When parsing, it will attempt to parse the value using
    all of the provided formats (stopping on the first successful parse).

    #format_datetime() can take into account the #date_formats and #time_formats as well if the
    *partial* parameter is set to #True.
    """

    name: str
    reference_url: t.Optional[str] = None
    date_formats: t.List[date_format] = field(default_factory=list)
    time_formats: t.List[time_format] = field(default_factory=list)
    datetime_formats: t.List[datetime_format] = field(default_factory=list)

    def parse_date(self, s: str) -> datetime.date:
        if not self.date_formats:
            raise ValueError(f"{self.name} has no date formats")
        for fmt in self.date_formats:
            try:
                return fmt.parse_date(s)
            except ValueError:
                pass
        raise _formulate_parse_error(self.name, self.date_formats, s)

    def format_date(self, d: datetime.date) -> str:
        if not self.date_formats:
            raise ValueError(f"{self.name} has no date formats")
        return self.date_formats[0].format_date(d)

    def parse_datetime(self, s: str, partial: bool = False) -> datetime.datetime:
        if not self.datetime_formats:
            raise ValueError(f"{self.name} has no datetime formats")
        for fmt in self.datetime_formats:
            try:
                return fmt.parse_datetime(s)
            except ValueError:
                pass
        if partial:
            try:
                return datetime.datetime.combine(self.parse_date(s), datetime.time.min)
            except ValueError:
                pass
            try:
                return datetime.datetime.combine(datetime.date.min, self.parse_time(s))
            except ValueError:
                pass
        raise _formulate_parse_error(self.name, self.datetime_formats, s)

    def format_datetime(self, dt: datetime.datetime) -> str:
        if not self.datetime_formats:
            raise ValueError(f"{self.name} has no datetime formats")
        return self.datetime_formats[0].format_datetime(dt)

    def parse_time(self, s: str) -> datetime.time:
        if not self.time_formats:
            raise ValueError(f"{self.name} has no time formats")
        for fmt in self.time_formats:
            try:
                return fmt.parse_time(s)
            except ValueError:
                pass
        raise _formulate_parse_error(self.name, self.time_formats, s)

    def format_time(self, t: datetime.time) -> str:
        if not self.time_formats:
            raise ValueError(f"{self.name} has no time formats")
        return self.time_formats[0].format_time(t)


#: Datetime format for parsing the format produced by `java.time.OffsetDateTime.toString()`.
JAVA_OFFSET_DATETIME = format_set(
    name="Java OffsetDateTime",
    reference_url="https://docs.oracle.com/javase/8/docs/api/java/time/OffsetDateTime.html#toString--",
    datetime_formats=[
        datetime_format.compile(r"%Y-%m-%dT%H:%M(:%S(\.%f)?)?%z?", regex_mode=True),
    ],
)


#: Date/time formats for parsing ISO 8601 dates and times.
#:
#: Supported ISO 8601 date formats are (implemented using the `%Y%m%d` format options):
#:
#: * `YYYY`
#: * `YYYY-MM` or `YYYYMM`
#: * `YYYY-MM-DD` or `YYYYMMDD`
#:
#: Note: ISO Week / ISO Week and day are not currently supported.
#:
#: Supported time formats are (implemented using the `%H%M%S%f` format options)
#:
#: * `hh`
#: * `hh:mm` or `hhmm`
#: * `hh:mm:ss` or `hhmmss`
#: * `hh:mm:ss.SSSSSS` (up to 6 sub-second digits)
#:
#: Supported timezone offset formats are (implemented using the `%z` format option):
#:
#: * `Z` (UTC)
#: * `+HH:MM` or `-HH:MM`
#: * `+HHMM` or `-HHMM`
#: * `+HH` or `-HH`
#:
#: Datetimes are simply concatenations of the possible date, time and offset formats, where
#: the date and time are separated by a `T`.
ISO_8601 = format_set(
    name="ISO 8061",
    reference_url="https://en.wikipedia.org/wiki/ISO_8601",
    date_formats=[
        date_format.compile(r"%Y(-%m(-%d)?)?", regex_mode=True),
        date_format.compile(r"%Y(%m%d?)?", regex_mode=True),
    ],
    time_formats=[
        time_format.compile(r"%H(:%M(:%S(\.%f)?)?)?%z?", regex_mode=True),
        time_format.compile(r"%H(%M%S??)?%z?", regex_mode=True),
    ],
    datetime_formats=[],
)

# Construct the datetime_formats, combining each element from the time and date formats.
for _i in range(2):
    ISO_8601.datetime_formats.append(
        datetime_format.compile(
            ISO_8601.date_formats[_i].format_str + "T" + ISO_8601.time_formats[_i].format_str,
            regex_mode=True,
        )
    )
del _i



import datetime
import typing as t
from dataclasses import dataclass

from .re import MatchFullError, match_full, matchiter

if t.TYPE_CHECKING:
    import dateutil.relativedelta

SECONDS_PER_DAY = 3600 * 24
DAYS_PER_MONTH = 30.4375
DAYS_PER_YEAR = 365.25
MICROSECONDS_PER_SECOND = 1000.0 * 1000.0


@dataclass
class duration:
    """
    Represents an ISO 8601 duration.

    __A note about sub-second percision__

    As described [here](https://stackoverflow.com/a/50570660/791713), referencing the ISO 8601 specification,
    sub-second precision must be expressed as a fraction in the lowest-order component, i.e. seconds. However,
    for the reasons of backwards compatibility, as well as parallels to the Python #datetime.timedelta class, we
    only store an additional #microseconds component which is extracted from the fractional representation of the
    ISO 8601 duration string.
    """

    years: int = 0
    months: int = 0
    weeks: int = 0
    days: int = 0
    hours: int = 0
    minutes: int = 0
    seconds: int = 0
    microseconds: int = 0

    _fields = ["years", "months", "weeks", "days", "hours", "minutes", "seconds", "microseconds"]

    def __post_init__(self) -> None:
        for k in self._fields:
            if getattr(self, k) < 0:
                raise ValueError("{} cannot be negative".format(k))

    def __str__(self) -> str:
        value: t.Union[int, str]
        parts = []

        date_parts = []
        for value, char in [(self.years, "Y"), (self.months, "M"), (self.weeks, "W"), (self.days, "D")]:
            if value != 0:
                date_parts.append("{}{}".format(value, char))
        parts.append("".join(date_parts))

        # Format the seconds component.
        if self.microseconds > 0:
            seconds = str(self.seconds + self.microseconds / MICROSECONDS_PER_SECOND).replace(".", ",")
        else:
            seconds = str(self.seconds)

        time_parts = []
        for value, char in [(str(self.hours), "H"), (str(self.minutes), "M"), (seconds, "S")]:
            if value != "0":
                time_parts.append("{}{}".format(value, char))
        if time_parts:
            parts.append("".join(time_parts))

        if not parts:
            return "PT0S"

        return "P" + "T".join(parts)

    def total_seconds(self) -> float:
        """
        Computes the total number of seconds in this duration.
        """

        return (
            self.years * DAYS_PER_YEAR * SECONDS_PER_DAY
            + self.months * DAYS_PER_MONTH * SECONDS_PER_DAY
            + self.weeks * 7 * SECONDS_PER_DAY
            + self.days * SECONDS_PER_DAY
            + self.hours * 3600
            + self.minutes * 60
            + self.seconds
            + self.microseconds / MICROSECONDS_PER_SECOND
        )

    def as_timedelta(self) -> datetime.timedelta:
        """
        Returns the seconds represented by this duration as a #datetime.timedelta object. The arguments
        and keyword arguments are forwarded to the #total_seconds() method.
        """

        return datetime.timedelta(seconds=self.total_seconds())

    def as_relativedelta(self) -> "dateutil.relativedelta.relativedelta":
        """
        Converts the #duration object to a #dateutil.relativedelta.relativedelta object. Requires
        the `python-dateutil` module.
        """

        from dateutil.relativedelta import relativedelta

        return relativedelta(
            years=self.years,
            months=self.months,
            weeks=self.weeks,
            days=self.days,
            hours=self.hours,
            minutes=self.minutes,
            seconds=self.seconds,
            microseconds=self.microseconds,
        )

    @classmethod
    def parse(cls, s: str) -> "duration":
        """
        Parses an ISO 8601 duration string into a #duration object.

        Thanks to https://stackoverflow.com/a/35936407.
        See also https://en.wikipedia.org/wiki/ISO_8601#Durations
        """

        parts = s.split("T")
        if not s or s[0] != "P" or len(parts) > 2:
            raise ValueError("Not an ISO 8601 duration string: {!r}".format(s))

        part_one = parts[0][1:]
        part_two = parts[1] if len(parts) == 2 else ""

        fields = {}

        try:
            for number, unit in (x.groups() for x in match_full(r"(\d+)(D|W|M|Y)", part_one)):
                number = int(number)
                if unit == "Y":
                    fields["years"] = number
                elif unit == "M":
                    fields["months"] = number
                elif unit == "W":
                    fields["weeks"] = number
                elif unit == "D":
                    fields["days"] = number

            last_match_end_idx = 0
            for last_match_end_idx, (number, unit) in (
                (x.end(), x.groups()) for x in matchiter(r"(\d+)(H|M)", part_two)
            ):
                number = int(number)
                if unit == "H":
                    fields["hours"] = number
                elif unit == "M":
                    fields["minutes"] = number

            part_three = part_two[last_match_end_idx:]
            for number in (x.group(1) for x in match_full(r"(\d+(?:[,\.]\d+)?)S", part_three)):
                number = float(number.replace(",", "."))
                fields["seconds"] = number

        except MatchFullError:
            raise ValueError("Not an ISO 8601 duration string: {!r}".format(s))

        seconds, remainder = divmod(fields.get("seconds", 0.0), 1.0)
        fields["seconds"] = int(seconds)
        fields["microseconds"] = int(remainder * MICROSECONDS_PER_SECOND)

        return cls(**fields)



import datetime
import io
import re
import typing as t
from dataclasses import dataclass

from .options import DatetimeComponentType, FormatOptions, IFormatOption
from .re import Pattern

_T_datetime_format = t.TypeVar("_T_datetime_format", bound="_datetime_format")


@dataclass
class _datetime_format:
    format_str: str
    regex: "Pattern"
    seq: t.List[t.Union[str, IFormatOption]]

    @classmethod
    def compile(cls: t.Type[_T_datetime_format], format_str: str, regex_mode: bool = False) -> "_T_datetime_format":
        """
        Compiles a format string to a static regex representation for fast parsing.
        """

        combo_regex = io.StringIO()
        combo_regex.write("^")
        combo_sequence: t.List[t.Union[str, IFormatOption]] = []

        def append_seq(char: str) -> None:
            if combo_sequence and isinstance(combo_sequence[-1], str):
                combo_sequence[-1] += char
            else:
                combo_sequence.append(char)

        idx = 0
        while idx < len(format_str):
            if regex_mode and format_str[idx] in "().?":
                if format_str[idx] == "(":
                    combo_regex.write("(?:")
                else:
                    combo_regex.write(format_str[idx])
            elif regex_mode and format_str[idx] == "\\":
                # TODO(NiklasRosenstein): oob check
                idx += 1
                char = format_str[idx]
                combo_regex.write(re.escape(char))
                append_seq(char)
            elif format_str[idx] == "%" and len(format_str) > idx + 1:
                # TODO(NiklasRosenstein): Interpret %% as single %
                idx += 1
                char = format_str[idx]
                option = FormatOptions.get(char)
                if option is None:
                    raise ValueError(f"unknown format option %{char}")
                combo_regex.write("(" + option.regex + ")")
                combo_sequence.append(option)
            else:
                char = format_str[idx]
                combo_regex.write(re.escape(char))
                append_seq(char)
            idx += 1

        combo_regex.write("$")
        return cls(format_str, re.compile(combo_regex.getvalue()), combo_sequence)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.format_str!r})"


@dataclass
class date_format(_datetime_format):
    __repr__ = _datetime_format.__repr__

    def __post_init__(self) -> None:
        for item in self.seq:
            if isinstance(item, IFormatOption) and item.component.type != DatetimeComponentType.Date:
                raise ValueError(f"%{item.char} is an invalid format option for date_format")

    def parse_date(self, s: str) -> datetime.date:
        # TODO(NiklasRosenstein): Validate that the format string actually only captures date components.
        dt = datetime_format(self.format_str, self.regex, self.seq).parse_datetime(s)
        return dt.date()

    def format_date(self, d: datetime.date) -> str:
        # TODO(NiklasRosenstein): Validate that the format string actually only captures date components.
        dt = datetime.datetime(d.year, d.month, d.day)
        return datetime_format(self.format_str, self.regex, self.seq).format_datetime(dt)

    def to_datetime_format(self) -> "datetime_format":
        return datetime_format(self.format_str, self.regex, self.seq)


@dataclass
class time_format(_datetime_format):
    __repr__ = _datetime_format.__repr__

    def __post_init__(self) -> None:
        for item in self.seq:
            if isinstance(item, IFormatOption) and item.component.type != DatetimeComponentType.Time:
                raise ValueError(f"%{item.char} is an invalid format option for time_format")

    def parse_time(self, s: str) -> datetime.time:
        # TODO(NiklasRosenstein): Validate that the format string actually only captures time components.
        dt = datetime_format(self.format_str, self.regex, self.seq).parse_datetime(s)
        return dt.time()

    def format_time(self, t: datetime.time) -> str:
        # TODO(NiklasRosenstein): Validate that the format string actually only captures time components.
        dt = datetime.datetime(1970, 1, 1, t.hour, t.minute, t.second, t.microsecond, t.tzinfo)
        return datetime_format(self.format_str, self.regex, self.seq).format_datetime(dt)

    def to_datetime_format(self) -> "datetime_format":
        return datetime_format(self.format_str, self.regex, self.seq)


class datetime_format(_datetime_format):
    def parse_datetime(self, s: str) -> datetime.datetime:
        match = self.regex.match(s)
        if not match:
            raise ValueError(f'"{s}" does not match format "{self.format_str}"')
        kwargs = {"year": 1900, "month": 1, "day": 1, "hour": 0}
        groups = iter(match.groups())
        for item in self.seq:
            if isinstance(item, IFormatOption):
                matched_string = next(groups)
                if matched_string is not None:
                    kwargs[item.component.value] = item.parse_string(matched_string)
        return datetime.datetime(**kwargs)  # type: ignore

    def format_datetime(self, dt: datetime.datetime) -> str:
        result = io.StringIO()
        for item in self.seq:
            if isinstance(item, str):
                result.write(item)
            else:
                result.write(item.format_value(dt, getattr(dt, item.component.value)))
        return result.getvalue()



__version__ = "2.1.0"

__all__ = [
    "tzlocal",
    "duration",
    "date_format",
    "datetime_format",
    "time_format",
    "parse_date",
    "parse_datetime",
    "parse_time",
    "format_date",
    "format_datetime",
    "format_date",
    "format_set",
    "JAVA_OFFSET_DATETIME",
    "ISO_8601",
]

import datetime
import time
import typing as t

from .duration import duration
from .format import date_format, datetime_format, time_format
from .format_sets import ISO_8601, JAVA_OFFSET_DATETIME, format_set


def tzlocal() -> datetime.tzinfo:
    offset = time.altzone if time.daylight else time.timezone
    return datetime.timezone(datetime.timedelta(seconds=-offset))


def parse_date(fmt: t.Union[format_set, date_format, str], s: str) -> datetime.date:
    if isinstance(fmt, str):
        fmt = date_format.compile(fmt)
    return fmt.parse_date(s)


def parse_datetime(fmt: t.Union[format_set, datetime_format, str], s: str) -> datetime.datetime:
    if isinstance(fmt, str):
        fmt = datetime_format.compile(fmt)
    return fmt.parse_datetime(s)


def parse_time(fmt: t.Union[format_set, time_format, str], s: str) -> datetime.time:
    if isinstance(fmt, str):
        fmt = time_format.compile(fmt)
    return fmt.parse_time(s)


def format_date(fmt: t.Union[format_set, date_format, str], d: datetime.date) -> str:
    if isinstance(fmt, str):
        fmt = date_format.compile(fmt)
    return fmt.format_date(d)


def format_datetime(fmt: t.Union[format_set, datetime_format, str], dt: datetime.datetime) -> str:
    if isinstance(fmt, str):
        fmt = datetime_format.compile(fmt)
    return fmt.format_datetime(dt)


def format_time(fmt: t.Union[format_set, time_format, str], t: datetime.time) -> str:
    if isinstance(fmt, str):
        fmt = time_format.compile(fmt)
    return fmt.format_time(t)



import abc
import datetime
import enum
import re
import typing as t
from dataclasses import dataclass


class DatetimeComponentType(enum.Enum):
    Date = enum.auto()
    Time = enum.auto()


class DatetimeComponent(enum.Enum):
    Year = "year"
    Month = "month"
    Day = "day"
    Hour = "hour"
    Minute = "minute"
    Second = "second"
    Microsecond = "microsecond"
    Timezone = "tzinfo"

    @property
    def type(self) -> DatetimeComponentType:
        if self in (DatetimeComponent.Year, DatetimeComponent.Month, DatetimeComponent.Day):
            return DatetimeComponentType.Date
        if self in (
            DatetimeComponent.Hour,
            DatetimeComponent.Minute,
            DatetimeComponent.Second,
            DatetimeComponent.Microsecond,
            DatetimeComponent.Timezone,
        ):
            return DatetimeComponentType.Time
        raise RuntimeError(f"unexpected Component enumeration value: {self!r}")


@dataclass
class _FormatOption:
    """
    Represents a special character in a date format string, which in turn represents a particular
    date or time component.
    """

    #: The character that the format option is represented in a format string. Must be a string of
    #: length 1 (one).
    char: str

    #: The component of a datetime construct that the format option represents.
    component: DatetimeComponent

    #: The regular expression that is used to capture the value represented by the format option
    #: when parsing a date/time string. This expression must not contain any capturing groups.
    regex: str


class IFormatOption(_FormatOption, metaclass=abc.ABCMeta):
    """
    Abstract metaclass on top of #_FormatOption, adding methods to the interface for parsing and
    formatting values represented by the format option.
    """

    @abc.abstractmethod
    def parse_string(self, s: str) -> t.Any:
        """Called to parse then value extracted by the *regex*."""

    @abc.abstractmethod
    def format_value(self, dt: datetime.datetime, v: t.Any) -> str:
        """Called to format the value of the respective #Component to a string."""


@dataclass
class NumericFormatOption(IFormatOption):
    """
    Default implementation for format options that represent a plain number component in the date.
    """

    format: t.Callable[[int], str]
    post_parse: t.Optional[t.Callable[[str], int]] = None

    def parse_string(self, s: str) -> t.Any:
        if self.post_parse is not None:
            return self.post_parse(s)
        return int(s)

    def format_value(self, dt: datetime.datetime, v: t.Any) -> str:
        assert isinstance(v, int), f"expected int to {self!r}.format_value()"
        return self.format(v)


@dataclass
class TimezoneFormatOption(IFormatOption):
    regex: str = r"(?:Z|[-+]\d{2}(?::?\d{2})?)"

    def parse_string(self, s: str) -> datetime.tzinfo:
        match = re.match(self.regex, s)
        if not match:
            raise ValueError("not a timezone string: {!r}".format(s))
        if s == "Z":
            return datetime.timezone.utc
        else:
            s = s.replace(":", "")
            sign = -1 if s[0] == "-" else 1
            hours = int(s[1:3])
            minutes = int(s[3:5] or "00")
            seconds = sign * (hours * 3600 + minutes * 60)
            return datetime.timezone(datetime.timedelta(seconds=seconds))

    def format_value(self, dt: datetime.datetime, v: t.Any) -> str:
        assert v is None or isinstance(v, datetime.tzinfo), f"expected datetime.tzinfo, got {v!r}"
        if v is None:
            return ""
        elif v == datetime.timezone.utc:
            return "Z"
        else:
            assert isinstance(v, datetime.tzinfo)
            utcoffset = v.utcoffset(dt)
            # NOTE Copied from CPython 3.7 datetime.py _format_offset()
            string = ""
            if utcoffset is not None:
                if utcoffset.days < 0:
                    sign = "-"
                    utcoffset = -utcoffset
                else:
                    sign = "+"
                off = utcoffset.total_seconds()
                hh, mm = divmod(off, 60 * 60)
                mm, ss = divmod(mm, 60)
                ss, ms = divmod(ss, 1)
                string += "%s%02d:%02d" % (sign, hh, mm)
                if ss or ms:
                    string += ":%02d" % ss
                    if ms:
                        string += ".%06d" % ms
            return string


class FormatOptions(enum.Enum):
    """
    Enumeration of all the available format options.
    """

    Year = NumericFormatOption("Y", DatetimeComponent.Year, r"\d{4}", lambda v: str(v).rjust(4, "0"))
    Month = NumericFormatOption("m", DatetimeComponent.Month, r"\d{2}", lambda v: str(v).rjust(2, "0"))
    Day = NumericFormatOption("d", DatetimeComponent.Day, r"\d{2}", lambda v: str(v).rjust(2, "0"))
    Hour = NumericFormatOption("H", DatetimeComponent.Hour, r"\d{2}", lambda v: str(v).rjust(2, "0"))
    Minute = NumericFormatOption("M", DatetimeComponent.Minute, r"\d{2}", lambda v: str(v).rjust(2, "0"))
    Second = NumericFormatOption("S", DatetimeComponent.Second, r"\d{2}", lambda v: str(v).rjust(2, "0"))
    Microsecond = NumericFormatOption(
        "f",
        DatetimeComponent.Microsecond,
        r"\d+",
        format=lambda v: str(v).rjust(6, "0").rstrip("0") or "0",
        post_parse=lambda v: int(str(int(v) * (10 ** max(6 - len(v), 0)))[:6]),
    )
    Timezone = TimezoneFormatOption("z", DatetimeComponent.Timezone)

    @classmethod
    def get(cls, char: str) -> t.Optional[IFormatOption]:
        # TODO(NiklasRosenstein): Room for optimization by caching available format options in dict.
        return next((x.value for x in cls if x.value.char == char), None)
